"""Tests for ComicInfo Tags."""

from datetime import date
from pathlib import Path

import pytest
from lxml import etree

from darkseid.comicinfo import ComicInfo
from darkseid.metadata import Arc, Basic, Credit, Metadata, Role, Series

CI_XSD = Path("tests/test_files/ComicInfo.xsd")


@pytest.fixture()
def test_credits() -> list[Credit]:
    return [
        Credit("Peter David", [Role("Writer")]),
        Credit("Martin Egeland", [Role("Penciller")]),
        Credit("Martin Egeland", [Role("Cover")]),
        Credit("Kevin Dooley", [Role("Editor")]),
        Credit("Howard Shum", [Role("Inker")]),
        Credit("Howard Shum", [Role("Cover")]),
        Credit("Tom McCraw", [Role("Colorist")]),
        Credit("Dan Nakrosis", [Role("Letterer")]),
    ]


@pytest.fixture()
def test_meta_data(test_credits: list[Credit]) -> Metadata:
    md = Metadata()
    md.publisher = Basic("DC Comics", 1)
    md.imprint = Basic("DC Black Label", 2)
    md.series = Series(
        "Aquaman",
        sort_name="Aquaman",
        volume=3,
        format="Annual",
        language="en",
    )
    md.issue = "1"
    md.stories = [Basic("Foo"), Basic("Bar")]
    md.cover_date = date(1993, 4, 15)
    md.characters = [
        Basic("Aquaman"),
        Basic("Mera"),
        Basic("Garth"),
    ]
    md.teams = [Basic("Atlanteans"), Basic("Justice League")]
    md.locations = [Basic("Atlantis"), Basic("Metropolis")]
    md.genres = [Basic("Super-Hero")]
    md.story_arcs = [
        Arc("Crisis on Infinite Earths"),
        Arc("Death of Aquagirl"),
    ]
    md.black_and_white = True
    md.age_rating = "MA15+"
    md.manga = "YesAndRightToLeft"
    for c in test_credits:
        md.add_credit(c)
    return md


def validate(xml_path: str, xsd_path: str) -> bool:
    xmlschema_doc = etree.parse(xsd_path)
    xmlschema = etree.XMLSchema(xmlschema_doc)

    xml_doc = etree.parse(xml_path)
    return xmlschema.validate(xml_doc)


def test_metadata_from_xml(test_meta_data: Metadata) -> None:
    """Simple test of creating the ComicInfo."""
    res = ComicInfo().string_from_metadata(test_meta_data)
    # TODO: add more asserts to verify data.
    assert res is not None


def test_meta_with_missing_stories(test_meta_data: Metadata, tmp_path: Path) -> None:
    """Test of writing the metadata to a file."""
    tmp_file = tmp_path / "test-write.xml"
    old_md = test_meta_data
    old_md.stories = None
    ComicInfo().write_to_external_file(tmp_file, test_meta_data)
    assert tmp_file.read_text() is not None
    assert validate(tmp_file, CI_XSD) is True
    new_md = ComicInfo().read_from_external_file(tmp_file)
    assert old_md.stories == new_md.stories
    assert old_md.characters == new_md.characters


def test_meta_write_to_file(test_meta_data: Metadata, tmp_path: Path) -> None:
    """Test of writing the metadata to a file."""
    tmp_file = tmp_path / "test-write.xml"
    ComicInfo().write_to_external_file(tmp_file, test_meta_data)
    assert tmp_file.read_text() is not None
    assert validate(tmp_file, CI_XSD) is True


def test_meta_write_to_existing_file(test_meta_data: Metadata, tmp_path: Path) -> None:
    # sourcery skip: extract-duplicate-method
    """Test of writing the metadata to a file and then modifying comicinfo.xml"""
    # Write test metadata to file
    tmp_file = tmp_path / "test-write.xml"
    ci = ComicInfo()
    ci.write_to_external_file(tmp_file, test_meta_data)
    assert tmp_file.read_text() is not None
    assert validate(tmp_file, CI_XSD) is True
    # Read the comicinfo.xml file and verify content
    md = ci.read_from_external_file(tmp_file)
    assert md.genres == test_meta_data.genres
    # Modify the metadata and overwrite the existing comicinfo.xml
    md.genres = []
    ci.write_to_external_file(tmp_file, md)
    assert tmp_file.read_text() is not None
    assert validate(tmp_file, CI_XSD) is True
    # Now reback the modified comicinfo.xml and verify
    new_md = ci.read_from_external_file(tmp_file)
    assert new_md.genres is None


def test_invalid_age_write_to_file(tmp_path: Path) -> None:
    """Test writing of invalid age rating value to a file."""
    aquaman = Series("Aquaman")
    bad_metadata = Metadata(series=aquaman, age_rating="MA 15+")
    tmp_file = tmp_path / "test-age-write.xml"
    ci = ComicInfo()
    ci.write_to_external_file(tmp_file, bad_metadata)
    result_md = ci.read_from_external_file(tmp_file)
    assert tmp_file.read_text() is not None
    assert validate(tmp_file, CI_XSD) is True
    assert result_md.age_rating == "Unknown"


def test_invalid_manga_write_to_file(tmp_path: Path) -> None:
    """Test writing of invalid manga value to a file."""
    aquaman = Series("Aquaman")
    bad_metadata = Metadata(series=aquaman, manga="Foo Bar")
    tmp_file = tmp_path / "test-manga-write.xml"
    ci = ComicInfo()
    ci.write_to_external_file(tmp_file, bad_metadata)
    result_md = ci.read_from_external_file(tmp_file)
    assert tmp_file.read_text() is not None
    assert validate(tmp_file, CI_XSD) is True
    assert result_md.manga == "Unknown"


def test_read_from_file(test_meta_data: Metadata, tmp_path: Path) -> None:
    """Test to read in the data from a file."""
    tmp_file = tmp_path / "test-read.xml"
    # Write metadata to file
    ComicInfo().write_to_external_file(tmp_file, test_meta_data)
    # Read the metadat from the file
    new_md = ComicInfo().read_from_external_file(tmp_file)

    assert new_md is not None
    assert new_md.series.name == test_meta_data.series.name
    assert new_md.series.language == new_md.series.language
    assert new_md.issue == test_meta_data.issue
    assert new_md.stories == test_meta_data.stories
    assert new_md.cover_date == test_meta_data.cover_date
    assert new_md.credits[0] == test_meta_data.credits[0]
    assert new_md.characters == test_meta_data.characters
    assert new_md.teams == test_meta_data.teams
    assert new_md.story_arcs == test_meta_data.story_arcs
    assert new_md.locations == test_meta_data.locations
    assert new_md.black_and_white == test_meta_data.black_and_white
    assert new_md.publisher.name == test_meta_data.publisher.name
    assert new_md.imprint.name == test_meta_data.imprint.name



from __future__ import annotations

import logging
import shutil
import tempfile
import zipfile
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path
from typing import cast

import rarfile

from darkseid.archivers.archiver import Archiver

logger = logging.getLogger(__name__)


class ZipArchiver(Archiver):
    """
    Handles archiving operations specific to ZIP files.

    This class provides methods for reading, writing, and removing files within a ZIP archive.
    """

    def __init__(self: ZipArchiver, path: Path) -> None:
        """
        Initializes a ZipArchiver object with the provided path.

        Args:
            path (Path): The path associated with the Zip file.

        Returns:
            None
        """
        super().__init__(path)

    def read_file(self: ZipArchiver, archive_file: str) -> bytes:
        """
        Reads the contents of a file from the ZIP archive.

        Args:
            archive_file (str): The file to read from the archive.

        Returns:
            bytes: The content of the file as bytes.

        Raises:
            OSError: If an error occurs during reading.
        """

        try:
            with zipfile.ZipFile(self.path, mode="r") as zf:
                return zf.read(archive_file)
        except (zipfile.BadZipfile, OSError) as e:
            logger.exception(
                "Error reading zip archive %s :: %s",
                self.path,
                archive_file,
            )
            raise OSError from e

    def remove_file(self: ZipArchiver, archive_file: str) -> bool:
        """
        Removes a file from the ZIP archive.

        Args:
            archive_file (str): The file to remove from the archive.

        Returns:
            bool: True if the file was successfully removed, False otherwise.
        """

        return self._rebuild([archive_file])

    def remove_files(self: ZipArchiver, filename_lst: list[str]) -> bool:
        """
        Removes multiple files from the ZIP archive.

        Args:
            filename_lst (list[str]): The list of files to remove from the archive.

        Returns:
            bool: True if all files were successfully removed, False otherwise.
        """

        return self._rebuild(filename_lst)

    def write_file(self: ZipArchiver, archive_file: str, data: str) -> bool:
        """
        Writes data to a file in the ZIP archive.

        Args:
            archive_file (str): The file to write to in the archive.
            data (str): The data to write to the file.

        Returns:
            bool: True if the write operation was successful, False otherwise.
        """

        #  At the moment, no other option but to rebuild the whole
        #  zip archive w/o the indicated file. Very sucky, but maybe
        # another solution can be found
        files = self.get_filename_list()
        if archive_file in files:
            self._rebuild([archive_file])

        try:
            # now just add the archive file as a new one
            with zipfile.ZipFile(
                self.path,
                mode="a",
                allowZip64=True,
                compression=zipfile.ZIP_DEFLATED,
            ) as zf:
                zf.writestr(archive_file, data)
        except (zipfile.BadZipfile, OSError):
            logger.exception(
                "Error writing zip archive %s :: %s",
                self.path,
                archive_file,
            )
            return False
        else:
            return True

    def get_filename_list(self: ZipArchiver) -> list[str]:
        """
        Returns a list of filenames in the ZIP archive.

        Returns:
            list[str]: A list of filenames in the archive.

        Raises:
            OSError: If an error occurs during retrieval.
        """

        try:
            with zipfile.ZipFile(self.path, mode="r") as zf:
                return zf.namelist()
        except (zipfile.BadZipfile, OSError):
            logger.exception("Error listing files in zip archive: %s", self.path)
            return []

    def _rebuild(self: ZipArchiver, exclude_list: list[str]) -> bool:
        """
        Rebuilds the ZIP archive excluding specified files.

        Args:
            exclude_list (list[str]): The list of files to exclude from the rebuild.

        Returns:
            bool: True if the rebuild was successful, False otherwise.
        """

        try:
            with zipfile.ZipFile(
                tempfile.NamedTemporaryFile(dir=self.path.parent, delete=False),
                "w",
                allowZip64=True,
            ) as zout:
                with zipfile.ZipFile(self.path, mode="r") as zin:
                    for item in zin.infolist():
                        buffer = zin.read(item.filename)
                        if item.filename not in exclude_list:
                            zout.writestr(item, buffer)

                # replace with the new file
                self.path.unlink(missing_ok=True)
                zout.close()  # Required on Windows
                shutil.move(cast(str, zout.filename), self.path)
        except (zipfile.BadZipfile, OSError):
            logger.exception("Error rebuilding zip file: %s", self.path)
            return False
        else:
            return True

    def copy_from_archive(self: ZipArchiver, other_archive: Archiver) -> bool:
        """
        Copies files from another archive to the ZIP archive.

        Args:
            other_archive (Archiver): The archive to copy files from.

        Returns:
            bool: True if the copy operation was successful, False otherwise.
        """

        try:
            with zipfile.ZipFile(self.path, mode="w", allowZip64=True) as zout:
                for filename in other_archive.get_filename_list():
                    try:
                        data = other_archive.read_file(filename)
                    except rarfile.BadRarFile:
                        # Skip any bad images in the file.
                        continue
                    if data is not None:
                        zout.writestr(filename, data)
        except (zipfile.BadZipfile, OSError) as e:
            # Remove any partial files created
            if self.path.exists():
                self.path.unlink()
            logger.warning("Error while copying to %s: %s", self.path, e)
            return False
        else:
            return True



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


class Archiver:
    """
    Handles archiving operations for files.

    This class provides methods for reading, writing, and removing files within an archive.
    """

    def __init__(self: Archiver, path: Path) -> None:
        """
        Initializes an Archiver object with a specified path.

        Args:
            path (Path): The path associated with the Archiver.

        Returns:
            None
        """

        self._path = path

    @property
    def path(self: Archiver) -> Path:
        """
        Returns the path associated with the Archiver.

        Returns:
            Path: The path associated with the Archiver.
        """

        return self._path

    def read_file(self: Archiver, archive_file: str) -> bytes:
        """
        Reads the content of a file from the archive.

        Args:
            archive_file (str): The file to read from the archive.

        Returns:
            bytes: The content of the file as bytes.

        Raises:
            NotImplementedError: Method or function hasn't been implemented yet.
        """

        raise NotImplementedError

    def write_file(
        self: Archiver,
        archive_file: str,  # noqa: ARG002
        data: str,  # noqa: ARG002
    ) -> bool:
        """
        Writes data to a file in the archive.

        Args:
            archive_file (str): The file to write to in the archive.
            data (str): The data to write to the file.

        Returns:
            bool: True if the write operation was successful, False otherwise.
        """

        return False

    def remove_file(self: Archiver, archive_file: str) -> bool:  # noqa: ARG002
        """
        Removes a file from the archive.

        Args:
            archive_file (str): The file to remove from the archive.

        Returns:
            bool: True if the file was successfully removed, False otherwise.
        """

        return False

    def remove_files(
        self: Archiver,
        filename_lst: list[str],  # noqa: ARG002
    ) -> bool:
        """
        Removes multiple files from the archive.

        Args:
            filename_lst (list[str]): The list of files to remove from the archive.

        Returns:
            bool: True if all files were successfully removed, False otherwise.
        """

        return False

    def get_filename_list(self: Archiver) -> list[str]:
        """
        Returns an empty list of filenames from the archive.

        Returns:
            list[str]: An empty list of filenames.
        """

        return []

    def copy_from_archive(
        self: Archiver,
        other_archive: Archiver,  # noqa: ARG002
    ) -> bool:
        """
        Copies files from another archive.

        Args:
            other_archive (Archiver): The archive to copy files from.

        Returns:
            bool: True if the copy operation was successful, False otherwise.
        """

        return False



import pytest

from darkseid.archivers.rar import RarArchiver


@pytest.fixture()
def rar_archiver(tmp_path):
    # Arrange
    rar_path = tmp_path / "test.rar"
    rar_path.touch()
    return RarArchiver(rar_path)


def test_remove_file(rar_archiver):
    # Act
    result = rar_archiver.remove_file("file.txt")

    # Assert
    assert result is False


def test_remove_files(rar_archiver):
    # Act
    result = rar_archiver.remove_files(["file1.txt", "file2.txt"])

    # Assert
    assert result is False


def test_write_file(rar_archiver):
    # Act
    result = rar_archiver.write_file("file.txt", "data")

    # Assert
    assert result is False


def test_copy_from_archive(rar_archiver):
    # Act
    result = rar_archiver.copy_from_archive(None)  # type: ignore

    # Assert
    assert result is False



# ruff: noqa: C901, PLR0915
"""A class for internal metadata storage.

The goal of this class is to handle ALL the data that might come from various
tagging schemes and databases, such as Metron, ComicVine or GCD.  This makes conversion
possible, however lossy it might be

"""

# Copyright 2012-2014 Anthony Beville
# Copyright 2020 Brian Pepple
from __future__ import annotations

from dataclasses import dataclass, field, fields
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from datetime import date
    from decimal import Decimal
from typing import TypedDict

import pycountry

MAX_UPC = 17
MAX_ISBN = 13
COUNTRY_LEN = 2


class Validations:
    def __post_init__(self: Validations) -> None:
        """
        Run validation methods if declared.

        The validation method can be a simple check that raises ValueError or a transformation
        to the field value. The validation is performed by calling a function named:
        `validate_<field_name>(self, value, field) -> field.type`.

        Args:
            self: The instance of the Validations class.

        Returns:
            None

        Example:
            ```python
            validations = Validations()
            validations.__post_init__()
            ```
        """
        for name, field_ in self.__dataclass_fields__.items():
            if method := getattr(self, f"validate_{name}", None):
                setattr(self, name, method(getattr(self, name), field=field_))


class PageType:
    """
    Defines constants for different types of pages.

    This class provides a set of predefined page types for categorizing pages in a publication.
    """

    FrontCover = "FrontCover"
    InnerCover = "InnerCover"
    Roundup = "Roundup"
    Story = "Story"
    Advertisement = "Advertisement"
    Editorial = "Editorial"
    Letters = "Letters"
    Preview = "Preview"
    BackCover = "BackCover"
    Other = "Other"
    Deleted = "Deleted"


class ImageMetadata(TypedDict, total=False):
    """
    Defines the structure of ImageMetadata using TypedDict.

    This class represents the metadata associated with an image.
    """

    Type: str
    Bookmark: str
    DoublePage: bool
    Image: int | str
    ImageSize: str
    ImageHeight: str
    ImageWidth: str


@dataclass
class Price(Validations):
    """
    A data class representing a price with validations.

    Attributes:
        amount (Decimal): The amount associated with the price.
        country (str): The country associated with the price, defaults to "US".
    """

    amount: Decimal
    country: str = field(default="US")

    @staticmethod
    def validate_country(value: str, **_: any) -> str:
        """
        Validates a country value.

        If the value is None, it returns the default country code "US". Otherwise, it strips
        any leading or trailing whitespace from the value. If the value is empty after
        stripping, it raises a ValueError.

        If the length of the value is 2, it tries to find the country object using the alpha-2
        code. Otherwise, it tries to look up the country object using the value. If the country
        object is not found, it raises a ValueError.

        Args:
            value (str): The country value to validate.
            **_ (any): Additional keyword arguments (ignored).

        Returns:
            str: The validated country code.

        Raises:
            ValueError: Raised when the country code cannot be found or when no value is given for the country.
        """

        if value is None:
            return "US"
        value = value.strip()
        if not value:
            msg = "No value given for country"
            raise ValueError(msg)

        if len(value) == COUNTRY_LEN:
            obj = pycountry.countries.get(alpha_2=value)
        else:
            try:
                obj = pycountry.countries.lookup(value)
            except LookupError as e:
                msg = f"Couldn't find country for {value}"
                raise ValueError(msg) from e

        if obj is None:
            msg = f"Couldn't get country code for {value}"
            raise ValueError(msg)
        return obj.alpha_2


@dataclass
class Basic:
    """
    A data class representing basic information.

    Attributes:
        name (str): The name associated with the basic information.
        id_ (int | None): The ID associated with the basic information, defaults to None.
    """

    name: str
    id_: int | None = None


@dataclass
class Universe(Basic):
    """
    A data class representing a universe.

    Attributes:
        name (str): The name associated with the basic information.
        id_ (int | None): The ID associated with the basic information, defaults to None.
        designation (str | None): The designation of the universe, defaults to None.
    """

    designation: str | None = None


@dataclass
class Role(Basic):
    """
    A data class representing a role.

    Attributes:
        name (str): The name associated with the basic information.
        id_ (int | None): The ID associated with the basic information, defaults to None.
        primary (bool): Indicates if the role is primary, defaults to False.
    """

    primary: bool = False


@dataclass
class Series(Basic, Validations):
    """
    A data class representing a series with basic information and validations.

    Attributes:
        name (str): The name associated with the basic information.
        id_ (int | None): The ID associated with the basic information, defaults to None.
        sort_name (str | None): The sort name of the series, defaults to None.
        volume (int | None): The volume of the series, defaults to None.
        format (str | None): The format of the series, defaults to None.
        language (str | None): The 2-letter ISO code of the language, defaults to None.

    Static Methods:
        validate_language(value: str, **_: any) -> str | None: Validates a language value.
    """

    sort_name: str | None = None
    volume: int | None = None
    format: str | None = None
    language: str | None = None  # 2-letter iso code

    @staticmethod
    def validate_language(value: str, **_: any) -> str | None:
        """
        Validates a language value.

        If the value is empty, it returns None. Otherwise, it strips any leading or trailing
        whitespace from the value. If the length of the value is 2, it tries to find the
        language object using the alpha-2 code. Otherwise, it tries to look up the language
        object using the value. If the language object is not found, it raises a ValueError.

        Args:
            value (str): The language value to validate.
            **_ (any): Additional keyword arguments (ignored).

        Returns:
            Optional[str]: The validated language code, or None if the value is empty.

        Raises:
            ValueError: Raised when the language object cannot be found.
        """

        if not value:
            return None
        value = value.strip()

        if len(value) == COUNTRY_LEN:
            obj = pycountry.languages.get(alpha_2=value)
        else:
            try:
                obj = pycountry.languages.lookup(value)
            except LookupError as e:
                msg = f"Couldn't find language {value}"
                raise ValueError(msg) from e
        if obj is None:
            msg = f"Couldn't find language {value}"
            raise ValueError(msg)
        return obj.alpha_2


@dataclass
class Arc(Basic):
    """
    A data class representing an arc with basic information.

    Attributes:
        name (str): The name associated with the basic information.
        id_ (int | None): The ID associated with the basic information, defaults to None.
        number (int | None): The number of the arc, defaults to None.
    """

    number: int | None = None


@dataclass
class Credit:
    """
    A data class representing a creator credit.

    Attributes:
        person (str): The name of the person associated with the credit.
        role (list[Role]): The list of roles associated with the credit.
        id_ (int | None): The ID associated with the credit, defaults to None.
    """

    person: str
    role: list[Role]
    id_: int | None = None


@dataclass
class GTIN(Validations):
    """
    A data class representing a GTIN (Global Trade Item Number) with validations.

    Attributes:
        upc (int | None): The UPC (Universal Product Code) associated with the GTIN, defaults to None.
        isbn (int | None): The ISBN (International Standard Book Number) associated with the GTIN, defaults to None.

    Static Methods:
        validate_upc(value: int, **_: any) -> int | None: Validates a UPC (Universal Product Code) value.
        validate_isbn(value: int, **_: any) -> int | None: Validates an ISBN (International Standard Book Number) value.
    """

    upc: int | None = None
    isbn: int | None = None

    @staticmethod
    def validate_upc(value: int, **_: any) -> int | None:
        """
        Validates a UPC (Universal Product Code) value.

        If the value is None or not an instance of int, it returns None. Otherwise, it checks
        if the length of the UPC value is greater than the maximum allowed length. If it is, it
        raises a ValueError.

        Args:
            value (int): The UPC value to validate.
            **_ (any): Additional keyword arguments (ignored).

        Returns:
            Optional[int]: The validated UPC value, or None if the value is None or not an instance of int.

        Raises:
            ValueError: Raised when the length of the UPC value is greater than the maximum allowed length.
        """

        # sourcery skip: class-extract-method
        if value is None or not isinstance(value, int):
            return None

        int_str = str(value)
        if len(int_str) > MAX_UPC:
            msg = f"UPC has a length greater than {MAX_UPC}"
            raise ValueError(msg)

        return value

    @staticmethod
    def validate_isbn(value: int, **_: any) -> int | None:
        """
        Validates an ISBN (International Standard Book Number) value.

        If the value is None or not an instance of int, it returns None. Otherwise, it checks
        if the length of the ISBN value is greater than the maximum allowed length. If it is,
        it raises a ValueError.

        Args:
            value (int): The ISBN value to validate.
            **_ (any): Additional keyword arguments (ignored).

        Returns:
            Optional[int]: The validated ISBN value, or None if the value is None or not an instance of int.

        Raises:
            ValueError: Raised when the length of the ISBN value is greater than the maximum allowed length.
        """

        if value is None or not isinstance(value, int):
            return None

        int_str = str(value)
        if len(int_str) > MAX_ISBN:
            msg = f"ISBN has a length greater than {MAX_ISBN}"
            raise ValueError(msg)

        return value


@dataclass
class Metadata:
    """
    Represents metadata for a comic.

    Attributes:
        is_empty (bool): Indicates if the metadata is empty.
        tag_origin (Optional[str]): The origin of the tag.
        info_source (Optional[Basic]): The information source.
        series (Optional[Series]): The series information.
        issue (Optional[str]): The issue information.
        collection_title (Optional[str]): The title of the collection.
        stories (list[Basic]): The list of stories.
        publisher (Optional[Basic]): The publisher information.
        imprint (Optional[Basic]): The imprint information.
        cover_date (Optional[date]): The cover date.
        store_date (Optional[date]): The store date.
        prices (list[Price]): The list of prices.
        gtin (Optional[GTIN]): The GTIN (Global Trade Item Number).
        issue_count (Optional[int]): The count of issues.
        genres (list[Basic]): The list of genres.
        comments (Optional[str]): The comments.
        volume_count (Optional[str]): The count of volumes.
        critical_rating (Optional[str]): The critical rating.
        country (Optional[str]): The country.
        alternate_series (Optional[str]): The alternate series.
        alternate_number (Optional[str]): The alternate number.
        alternate_count (Optional[int]): The count of alternates.
        notes (Optional[str]): The notes.
        web_link (Optional[str]): The web link.
        manga (Optional[str]): The manga information.
        black_and_white (Optional[bool]): Indicates if the comic is black and white.
        page_count (Optional[int]): The count of pages.
        age_rating (Optional[str]): The age rating.
        story_arcs (list[Arc]): The list of story arcs.
        series_group (Optional[str]): The series group.
        scan_info (Optional[str]): The scan information.
        characters (list[Basic]): The list of characters.
        teams (list[Basic]): The list of teams.
        locations (list[Basic]): The list of locations.
        credits (list[Credit]): The list of credits.
        reprints (list[Basic]): The list of reprints.
        tags (list[Basic]): The list of tags.
        pages (list[ImageMetadata]): The list of pages.

    Methods:
        __post_init__: Initializes the metadata object.
        overlay: Overlays a metadata object on this one.
        overlay_credits: Overlays the credits from a new metadata object.
        set_default_page_list: Sets a default page list.
        get_archive_page_index: Gets the archive page index for a given displayed page number.
        get_cover_page_index_list: Gets a list of archive page indices of cover pages.
        _existing_credit: Checks if a credit with the given creator already exists.
        _role_exists: Checks if a role already exists in the old roles list.
        add_credit: Adds a new credit to the metadata.
        __str__: Returns a string representation of the metadata.

    Example:
        ```python
        metadata = Metadata(is_empty=True)
        metadata.overlay(new_md)
        print(metadata)
        ```
    """

    is_empty: bool = True
    tag_origin: str | None = None

    info_source: Basic | None = None
    series: Series | None = None
    issue: str | None = None
    collection_title: str | None = None
    stories: list[Basic] = field(default_factory=list)
    publisher: Basic | None = None
    imprint: Basic | None = None
    cover_date: date | None = None
    store_date: date | None = None
    prices: list[Price] = field(default_factory=list)
    gtin: GTIN | None = None
    issue_count: int | None = None
    genres: list[Basic] = field(default_factory=list)
    comments: str | None = None  # use same way as Summary in CIX

    volume_count: str | None = None
    critical_rating: str | None = None
    country: str | None = None

    alternate_series: str | None = None
    alternate_number: str | None = None
    alternate_count: int | None = None
    notes: str | None = None
    web_link: str | None = None
    manga: str | None = None
    black_and_white: bool | None = None
    page_count: int | None = None
    age_rating: str | None = None

    story_arcs: list[Arc] = field(default_factory=list)
    series_group: str | None = None
    scan_info: str | None = None

    characters: list[Basic] = field(default_factory=list)
    teams: list[Basic] = field(default_factory=list)
    locations: list[Basic] = field(default_factory=list)
    universes: list[Universe] = field(default_factory=list)

    credits: list[Credit] = field(default_factory=list)
    reprints: list[Basic] = field(default_factory=list)
    tags: list[Basic] = field(default_factory=list)
    pages: list[ImageMetadata] = field(default_factory=list)

    def __post_init__(self: Metadata) -> None:
        """
        Executes the post-initialization process for a Metadata instance.

        The method iterates over the attributes of the Metadata object and checks if any
        attribute has a non-empty value, excluding the "is_empty" attribute. If a non-empty
        value is found, the "is_empty" attribute is set to False and the iteration is stopped.

        Args:
            self (Metadata): The Metadata instance.

        Returns:
            None

        Example:
            ```python
            metadata = Metadata()
            metadata.__post_init__()
            print(metadata.is_empty)  # Output: False
            ```
        """
        for key, value in self.__dict__.items():
            if value and key != "is_empty":
                self.is_empty = False
                break

    def overlay(self: Metadata, new_md: Metadata) -> None:  # noqa: PLR0912
        """
        Overlays a metadata object on this one.

        The method assigns non-None values from the new metadata object to the corresponding
        attributes of the current metadata object. If a value is an empty string, it is
        assigned as None. The "is_empty" attribute of the current metadata object is set to
        False if the new metadata object is not empty.

        Args:
            self (Metadata): The current metadata object.
            new_md (Metadata): The new metadata object to overlay.

        Returns:
            None

        Example:
            ```python
            metadata = Metadata()
            new_metadata = Metadata(series="Series 1", issue="Issue 1")
            metadata.overlay(new_metadata)
            print(metadata.series)  # Output: "Series 1"
            print(metadata.issue)  # Output: "Issue 1"
            ```
        """

        def assign(cur: str, new: any) -> None:
            if new is not None:
                if isinstance(new, str) and len(new) == 0:
                    setattr(self, cur, None)
                else:
                    setattr(self, cur, new)

        if not new_md.is_empty:
            self.is_empty = False

        assign("series", new_md.series)
        assign("info_source", new_md.info_source)
        assign("issue", new_md.issue)
        assign("collection_title", new_md.collection_title)
        assign("issue_count", new_md.issue_count)
        if len(new_md.stories) > 0:
            assign("stories", new_md.stories)
        assign("publisher", new_md.publisher)
        assign("cover_date", new_md.cover_date)
        assign("store_date", new_md.store_date)
        if len(new_md.prices) > 0:
            assign("price", new_md.prices)
        assign("gtin", new_md.gtin)
        assign("volume_count", new_md.volume_count)
        if len(new_md.genres) > 0:
            assign("genre", new_md.genres)
        assign("country", new_md.country)
        assign("critical_rating", new_md.critical_rating)
        assign("alternate_series", new_md.alternate_series)
        assign("alternate_number", new_md.alternate_number)
        assign("alternate_count", new_md.alternate_count)
        assign("imprint", new_md.imprint)
        assign("web_link", new_md.web_link)
        assign("manga", new_md.manga)
        assign("black_and_white", new_md.black_and_white)
        assign("age_rating", new_md.age_rating)
        if len(new_md.story_arcs) > 0:
            assign("story_arcs", new_md.story_arcs)
        assign("series_group", new_md.series_group)
        assign("scan_info", new_md.scan_info)
        if len(new_md.characters) > 0:
            assign("characters", new_md.characters)
        if len(new_md.teams) > 0:
            assign("teams", new_md.teams)
        if len(new_md.locations) > 0:
            assign("locations", new_md.locations)
        if len(new_md.universes) > 0:
            assign("universes", new_md.universes)
        if len(new_md.reprints) > 0:
            assign("reprints", new_md.reprints)
        assign("comments", new_md.comments)
        assign("notes", new_md.notes)

        if new_md.credits:
            self.overlay_credits(new_md.credits)

        # For now, go the easy route, where any overlay
        # value wipes out the whole list
        if len(new_md.tags) > 0:
            assign("tags", new_md.tags)

        if len(new_md.pages) > 0:
            assign("pages", new_md.pages)

    def overlay_credits(self: Metadata, new_credits: list[Credit]) -> None:
        """
        Overlays the credits from a new metadata object on the current metadata object.

        The method iterates over the new credits and removes any credit role if the person is
        blank. If the person is not blank, the "primary" attribute of the credit is set based
        on the presence of the "primary" key in the credit dictionary. The credit is then
        added to the current metadata object using the "add_credit" method.

        Args:
            self (Metadata): The current metadata object.
            new_credits (list[Credit]): The list of new credits to overlay.

        Returns:
            None

        Example:
            ```python
            metadata = Metadata()
            new_credits = [Credit(person="John Doe", role="Writer")]
            metadata.overlay_credits(new_credits)
            print(metadata.credits)  # Output: [Credit(person="John Doe", role="Writer")]
            ```
        """
        for credit in new_credits:
            # Remove credit role if person is blank
            if credit.person == "":
                for r in reversed(self.credits):
                    if r.role.casefold() == credit.role.casefold():
                        self.credits.remove(r)
            else:
                credit.primary = bool("primary" in credit and credit.primary)
                self.add_credit(credit)

    def set_default_page_list(self: Metadata, count: int) -> None:
        """
        Generates a default page list for the Metadata object.

        The method creates a default page list with the specified count. Each page is
        represented by an ImageMetadata object with the "Image" attribute set to the
        corresponding index. The first page is marked as the front cover by setting the "Type"
        attribute to PageType.FrontCover. The generated page list is appended to the "pages"
        attribute of the Metadata object.

        Args:
            self (Metadata): The Metadata object.
            count (int): The count of pages to generate.

        Returns:
            None

        Example:
            ```python
            metadata = Metadata()
            metadata.set_default_page_list(5)
            print(metadata.pages)
            # Output: [ImageMetadata(Image=0, Type=PageType.FrontCover),
            # ImageMetadata(Image=1), ImageMetadata(Image=2), ImageMetadata(Image=3),
            # ImageMetadata(Image=4)]
            ```
        """
        # generate a default page list, with the first page marked as the cover
        for i in range(count):
            page_dict = ImageMetadata(Image=i)
            if i == 0:
                page_dict["Type"] = PageType.FrontCover
            self.pages.append(page_dict)

    def get_archive_page_index(self: Metadata, pagenum: int) -> int:
        """
        Converts the displayed page number to the page index of the file in the archive.

        The method takes a displayed page number and returns the corresponding page index in
        the archive. If the displayed page number is within the range of the available pages,
        the "Image" attribute of the corresponding page in the "pages" attribute of the
        Metadata object is returned as an integer. If the displayed page number is out of
        range, 0 is returned.

        Args:
            self (Metadata): The Metadata object.
            pagenum (int): The displayed page number.

        Returns:
            int: The page index in the archive.

        Example:
            ```python
            metadata = Metadata()
            metadata.pages = [ImageMetadata(Image=0), ImageMetadata(Image=1), ImageMetadata(Image=2)]
            index = metadata.get_archive_page_index(1)
            print(index)  # Output: 1
            ```
        """

        # convert the displayed page number to the page index of the file in the archive
        return int(self.pages[pagenum]["Image"]) if pagenum < len(self.pages) else 0

    def get_cover_page_index_list(self: Metadata) -> list[int]:
        """
        Returns a list of archive page indices of cover pages.

        The method iterates over the pages in the Metadata object and checks if a page is
        marked as a front cover by having a "Type" attribute equal to PageType.FrontCover. If a
        front cover page is found, its "Image" attribute is added to the coverlist. If no front
        cover pages are found, the coverlist is initialized with a single element of 0.

        Returns:
            list[int]: The list of archive page indices of cover pages.

        Example:
            ```python
            metadata = Metadata()
            metadata.pages = [ImageMetadata(Image=0, Type=PageType.FrontCover),ImageMetadata(Image=1), ImageMetadata(Image=2)]
            cover_indices = metadata.get_cover_page_index_list()
            print(cover_indices)  # Output: [0]
            ```
        """
        coverlist = [
            int(p["Image"]) for p in self.pages if "Type" in p and p["Type"] == PageType.FrontCover
        ]

        if not coverlist:
            coverlist.append(0)

        return coverlist

    def _existing_credit(self: Metadata, creator: str) -> tuple[bool, int | None]:
        """
        Checks if a credit with the specified creator already exists in the Metadata.

        Args:
            creator (str): The creator to check for in the existing credits.

        Returns:
            tuple[bool, int | None]: A tuple containing a boolean indicating if the credit exists and the index of the existing credit, or None if not found.
        """

        return (
            next(
                (
                    (True, i)
                    for i, existing in enumerate(self.credits)
                    if creator.casefold() == existing.person.casefold()
                ),
                (False, None),
            )
            if self.credits
            else (False, None)
        )

    @staticmethod
    def _role_exists(new_role: Role, old_roles: list[Role]) -> bool:
        """
        Checks if a role already exists in the list of old roles.

        Args:
            new_role (Role): The new role to check for in the old roles list.
            old_roles (list[Role]): The list of old roles to check against.

        Returns:
            bool: True if the role already exists, False otherwise.
        """

        return any(role.name.casefold() == new_role.name.casefold() for role in old_roles)

    def add_credit(self: Metadata, new_credit: Credit) -> None:
        """
        Adds a new credit to the Metadata.

        If a credit with the same person already exists, the roles from the new credit are added to the existing credit.
        If the person is new, the new credit is appended to the list of credits.

        Args:
            new_credit (Credit): The new credit to add to the Metadata.

        Returns:
            None
        """

        exist, idx = self._existing_credit(new_credit.person)
        if exist:
            existing_credit: Credit = self.credits[idx]
            for new_role in new_credit.role:
                if not self._role_exists(new_role, existing_credit.role):
                    existing_credit.role.append(new_role)
        else:
            self.credits.append(new_credit)

    def __str__(self: Metadata) -> str:
        """
        Returns a string representation of the Metadata object with its attributes and values.
        """

        cls = self.__class__
        cls_name = cls.__name__
        indent = " " * 4
        res = [f"{cls_name}("]
        for f in fields(cls):
            value = getattr(self, f.name)
            if value is not None:
                res.append(f"{indent}{f.name} = {value!r},")
        res.append(")")
        return "\n".join(res)



import pytest

from darkseid.archivers.__init__ import UnknownArchiver


@pytest.mark.parametrize("expected_name", ["Unknown"], ids=["happy_path"])
def test_name(expected_name):
    # Act
    result = UnknownArchiver.name()

    # Assert
    assert result == expected_name



from pathlib import Path

import pytest

from darkseid import utils

test_articles = [
    pytest.param("The Champions & Inhumans", "Test string with '&'", "Champions Inhumans"),
    pytest.param("Justice League", "Test string with no articles", "Justice League"),
    pytest.param("The X-Men", "Test string with leading 'The'", "X-Men"),
]


@pytest.mark.parametrize(("test_string", "reason", "expected"), test_articles)
def test_file_name_for_articles(
    test_string: str,
    reason: str,  # noqa: ARG001
    expected: str,
) -> None:
    result = utils.remove_articles(test_string)

    assert result == expected


test_string_lists = [
    pytest.param(["apple", "banana", "cherry"], "Normal string list", "apple, banana, cherry"),
    pytest.param(
        ["Outsiders", "Infinity, Inc.", "Teen Titans"],
        "String list with comma value",
        'Outsiders, "Infinity, Inc.", Teen Titans',
    ),
]


@pytest.mark.parametrize(("test_list", "reason", "expected"), test_string_lists)
def test_list_to_string(test_list: list[str], reason: str, expected: str) -> None:  # noqa: ARG001
    result = utils.list_to_string(test_list)
    assert result == expected


def test_unique_name(tmp_path: Path) -> None:
    d = tmp_path / "test"
    d.mkdir()
    new_file = d / "Test.cbz"
    new_file.write_text("Blah")
    result = utils.unique_file(new_file)
    expected_result = Path(new_file.parent / f"{new_file.stem} (1){new_file.suffix}")

    assert result == expected_result


def test_recursive_list_with_file(tmp_path: Path) -> None:
    temp_file = tmp_path / "test.cbz"
    temp_file.write_text("foo")
    temp_cbr = tmp_path / "fugazi.cbr"
    temp_cbr.write_text("You are not what you own")
    # The following file should be *excluded* from results
    temp_txt = tmp_path / "fail.txt"
    temp_txt.write_text("yikes")

    expected_result = [temp_cbr, temp_file]
    result = utils.get_recursive_filelist([tmp_path])

    assert result == expected_result


def test_recursive_list_with_directory(tmp_path: Path) -> None:
    temp_dir = tmp_path / "recursive"
    temp_dir.mkdir()

    # Not really zip files, but we really only care about the file extension
    temp_file_1 = temp_dir / "test-1.cbz"
    temp_file_1.write_text("content")

    temp_file_2 = temp_dir / "test-2.cbz"
    temp_file_2.write_text("content")

    expected_result = [temp_file_2, temp_file_1]
    expected_result = sorted(expected_result)

    file_list = [temp_dir]
    result = utils.get_recursive_filelist(file_list)

    assert result == expected_result



"""A class to represent a single comic."""

# Copyright 2012-2014 Anthony Beville
# Copyright 2019 Brian Pepple
from __future__ import annotations

import io
import logging
import os
import zipfile
from pathlib import Path

import rarfile
from natsort import natsorted, ns
from PIL import Image

from darkseid.archivers import UnknownArchiver
from darkseid.archivers.rar import RarArchiver
from darkseid.archivers.zip import ZipArchiver
from darkseid.comicinfo import ComicInfo
from darkseid.metadata import Metadata

logger = logging.getLogger(__name__)


class Comic:
    """
    The Comic class represents a comic object with methods for interacting with comic archives.
    """

    class ArchiveType:
        """
        Types of archives supported.

        The ArchiveType class defines the types of archives supported, including zip, rar, and unknown.
        """

        zip, rar, unknown = list(range(3))  # noqa: RUF012

    def __init__(self: Comic, path: Path | str) -> None:
        """
        Initializes a Comic object with the provided path.

        Args:
            path (Path | str): The path to the comic file.

        Returns:
            None
        """
        self._path: Path = Path(path) if isinstance(path, str) else path
        self._ci_xml_filename: str = "ComicInfo.xml"
        self._has_md: bool | None = None
        self._page_count: int | None = None
        self._page_list: list[str] | None = None
        self._metadata: Metadata | None = None

        if self.zip_test():
            self._archive_type: int = self.ArchiveType.zip
            self._archiver = ZipArchiver(self._path)
        elif self.rar_test():
            self._archive_type: int = self.ArchiveType.rar
            self._archiver = RarArchiver(self._path)
        else:
            self._archive_type = self.ArchiveType.unknown
            self._archiver = UnknownArchiver(self._path)

    def __str__(self: Comic) -> str:
        """
        Returns the name of the comic file.
        """

        return f"{self._path.name}"

    @property
    def path(self: Comic) -> Path:
        """
        Returns the path of the comic.

        Returns:
            Path: The path of the comic.
        """

        return self._path

    @property
    def archiver(self: Comic) -> RarArchiver | ZipArchiver | UnknownArchiver:
        """
        Returns the archiver used for the comic.

        Returns:
            RarArchiver | ZipArchiver | UnknownArchiver: The archiver object used for the comic.
        """

        return self._archiver

    def reset_cache(self: Comic) -> None:
        """
        Clears the cached data.
        """

        self._has_md = None
        self._page_count = None
        self._page_list = None
        self._metadata = None

    def rar_test(self: Comic) -> bool:
        """
        Tests whether the provided path is a rar file.

        Returns:
            bool: True if the path is a rar file, False otherwise.
        """

        return rarfile.is_rarfile(self._path)

    def zip_test(self: Comic) -> bool:
        """
        Tests whether the provided path is a zipfile.

        Returns:
            bool: True if the path is a zipfile, False otherwise.
        """

        return zipfile.is_zipfile(self._path)

    def is_rar(self: Comic) -> bool:
        """
        Returns a boolean indicating whether the archive is a rarfile.
        """

        return self._archive_type == self.ArchiveType.rar

    def is_zip(self: Comic) -> bool:
        """
        Returns a boolean indicating whether the archive is a zipfile.
        """

        return self._archive_type == self.ArchiveType.zip

    def is_writable(self: Comic) -> bool:
        """
        Returns a boolean indicating whether the archive is writable.

        Returns:
            bool: True if the archive is writable, False otherwise.
        """

        if self._archive_type in [self.ArchiveType.unknown, self.ArchiveType.rar]:
            return False

        return bool(os.access(self._path, os.W_OK))

    def seems_to_be_a_comic_archive(self: Comic) -> bool:
        """
        Returns a boolean indicating whether the file is a comic archive.
        """

        return bool(
            (self.is_zip() or self.is_rar()) and (self.get_number_of_pages() > 0),
        )

    def get_page(self: Comic, index: int) -> bytes | None:
        """
        Returns an image(page) from an archive.

        Args:
            index (int): The index of the page to retrieve.

        Returns:
            bytes | None: The image data of the page, or None if an error occurs.
        """

        image_data = None

        filename = self.get_page_name(index)

        if filename is not None:
            try:
                image_data = self._archiver.read_file(filename)
            except OSError:
                logger.exception("Error reading '%s' from '%s'", filename, self._path)

        return image_data

    def get_page_name(self: Comic, index: int) -> str | None:
        """
        Returns the page name from an index.

        Args:
            index (int): The index of the page.

        Returns:
            str | None: The name of the page, or None if the index is out of range.
        """

        if index is None:
            return None

        page_list = self.get_page_name_list()

        num_pages = len(page_list)
        return None if num_pages == 0 or index >= num_pages else page_list[index]

    @staticmethod
    def is_image(name_path: Path) -> bool:
        """
        Checks if the given path is an image file based on its suffix.

        Args:
            name_path (Path): The path to check.

        Returns:
            bool: True if the path is an image file, False otherwise.
        """

        suffix_list = [".jpg", ".jpeg", ".png", ".gif", ".webp"]
        return name_path.suffix.casefold() in suffix_list and name_path.name[0] != "."

    def get_page_name_list(self: Comic, sort_list: bool = True) -> list[str]:
        """
        Returns a list of page names from an archive.

        Args:
            sort_list (bool): Indicates whether to sort the list. Default is True.

        Returns:
            list[str]: A list of page names from the archive.
        """

        if self._page_list is None:
            # get the list file names in the archive, and sort
            files = self._archiver.get_filename_list()

            # seems like some archive creators are on  Windows, and don't know
            # about case-sensitivity!
            if sort_list:
                files = natsorted(files, alg=ns.IGNORECASE)

            # make a sub-list of image files
            self._page_list = []
            for name in files:
                name_str = str(name)
                if self.is_image(Path(name_str)):
                    self._page_list.append(name_str)

        return self._page_list

    def get_number_of_pages(self: Comic) -> int:
        """
        Returns the number of pages in an archive.
        """

        if self._page_count is None:
            self._page_count = len(self.get_page_name_list())
        return self._page_count

    def read_metadata(self: Comic) -> Metadata:
        """
        Reads the metadata from an archive if present.

        Returns:
            Metadata: The metadata object.
        """

        if self._metadata is None:
            raw_metadata = self.read_raw_metadata()
            if raw_metadata is None or raw_metadata == "":
                self._metadata = Metadata()
            else:
                self._metadata = ComicInfo().metadata_from_string(raw_metadata)

            # validate the existing page list (make sure count is correct)
            if len(self._metadata.pages) not in [0, self.get_number_of_pages()]:
                # pages array doesn't match the actual number of images we're seeing
                # in the archive, so discard the data
                self._metadata.pages = []

            if len(self._metadata.pages) == 0:
                self._metadata.set_default_page_list(self.get_number_of_pages())

        return self._metadata

    def read_raw_metadata(self: Comic) -> str | None:
        """
        Reads the raw metadata from the comic file.

        Returns:
            str | None: The raw metadata as a string, or None if metadata is not available or an error occurs.
        """

        if not self.has_metadata():
            return None
        try:
            tmp_raw_metadata = self._archiver.read_file(self._ci_xml_filename)
            # Convert bytes to str. Is it safe to decode with utf-8?
            raw_metadata = tmp_raw_metadata.decode("utf-8")
        except OSError:
            logger.exception("Error reading in raw CIX!")
            raw_metadata = None
        return raw_metadata

    def write_metadata(self: Comic, metadata: Metadata | None) -> bool:
        """
        Write the metadata to the archive.

        Args:
            metadata (Metadata | None): The metadata object to write.

        Returns:
            bool: True if the write operation was successful, False otherwise.
        """

        if metadata is None or not self.is_writable():
            return False
        self.apply_archive_info_to_metadata(metadata, calc_page_sizes=True)
        if raw_cix := self.read_raw_metadata():
            md_string = ComicInfo().string_from_metadata(metadata, raw_cix.encode("utf-8"))
        else:
            md_string = ComicInfo().string_from_metadata(metadata)
        write_success = self._archiver.write_file(self._ci_xml_filename, md_string)
        return self._successful_write(write_success, True, metadata)

    def remove_metadata(self: Comic) -> bool:
        """
        Remove the metadata from the archive if present.

        Returns:
            bool: True if the metadata was successfully removed, False otherwise.
        """

        if self.has_metadata():
            write_success = self._archiver.remove_file(self._ci_xml_filename)
            return self._successful_write(write_success, False, None)
        return True

    def remove_pages(self: Comic, pages_index: list[int]) -> bool:
        """
        Remove pages from the archive.

        Args:
            pages_index (list[int]): List of page indices to remove.

        Returns:
            bool: True if the pages were successfully removed, False otherwise.
        """

        if not pages_index:
            return False
        pages_name_lst = []
        for idx in pages_index:
            page_name = self.get_page_name(idx)
            pages_name_lst.append(page_name)
        write_success = self._archiver.remove_files(pages_name_lst)
        return self._successful_write(write_success, False, None)

    def _successful_write(
        self: Comic,
        write_success: bool,
        has_md: bool,
        metadata: Metadata | None,
    ) -> bool:
        """
        Updates the state based on the success of a write operation.

        Args:
            write_success (bool): Indicates if the write operation was successful.
            has_md (bool): Indicates if metadata is present.
            metadata (Metadata | None): The metadata object.

        Returns:
            bool: The success status of the write operation.
        """

        if write_success:
            self._has_md = has_md
            self._metadata = metadata
        self.reset_cache()
        return write_success

    def has_metadata(self: Comic) -> bool:
        """
        Checks to see if the archive has metadata.

        Returns:
            bool: True if the archive has metadata, False otherwise.
        """

        if self._has_md is None:
            self._has_md = bool(
                self.seems_to_be_a_comic_archive()
                and (
                    not self.seems_to_be_a_comic_archive()
                    or self._ci_xml_filename in self._archiver.get_filename_list()
                ),
            )

        return self._has_md

    def apply_archive_info_to_metadata(
        self: Comic,
        metadata: Metadata,
        calc_page_sizes: bool = False,
    ) -> None:
        """
        Apply page information from the archive to the metadata.

        Args:
            metadata (Metadata): The metadata object to update.
            calc_page_sizes (bool): Indicates whether to calculate page sizes. Default is False.

        Returns:
            None
        """

        metadata.page_count = self.get_number_of_pages()

        if calc_page_sizes:
            for page in metadata.pages:
                if "ImageSize" not in page or "ImageHeight" not in page or "ImageWidth" not in page:
                    idx = int(page["Image"])  # type: ignore
                    data = self.get_page(idx)
                    if data is not None:
                        try:
                            page_image = Image.open(io.BytesIO(data))
                            width, height = page_image.size

                            page["ImageSize"] = str(len(data))
                            page["ImageHeight"] = str(height)
                            page["ImageWidth"] = str(width)
                        except OSError:
                            page["ImageSize"] = str(len(data))

    def export_as_zip(self: Comic, zip_filename: Path) -> bool:
        """
        Export CBR archives to CBZ format.

        Args:
            zip_filename (Path): The filename for the zip archive.

        Returns:
            bool: True if the export operation was successful, False otherwise.
        """

        if self._archive_type == self.ArchiveType.zip:
            # nothing to do, we're already a zip
            return True

        zip_archiver = ZipArchiver(zip_filename)

        return zip_archiver.copy_from_archive(self._archiver)



from pathlib import Path

import pytest

from darkseid.archivers.archiver import Archiver


@pytest.fixture()
def archiver():
    return Archiver(Path("/fake/path"))


def test_init():
    # Arrange
    path = Path("/fake/path")

    # Act
    archiver = Archiver(path)

    # Assert
    assert archiver.path == path


@pytest.mark.parametrize(
    ("archive_file", "expected_exception"),
    [
        ("file1.txt", NotImplementedError),
        ("file2.txt", NotImplementedError),
    ],
    ids=["file1.txt", "file2.txt"],
)
def test_read_file(archiver, archive_file, expected_exception):
    # Act & Assert
    with pytest.raises(expected_exception):
        archiver.read_file(archive_file)


@pytest.mark.parametrize(
    ("archive_file", "data", "expected_result"),
    [
        ("file1.txt", "data1", False),
        ("file2.txt", "data2", False),
    ],
    ids=["file1.txt_data1", "file2.txt_data2"],
)
def test_write_file(archiver, archive_file, data, expected_result):
    # Act
    result = archiver.write_file(archive_file, data)

    # Assert
    assert result == expected_result


@pytest.mark.parametrize(
    ("archive_file", "expected_result"),
    [("file1.txt", False), ("file2.txt", False)],
    ids=["file1.txt", "file2.txt"],
)
def test_remove_file(archiver, archive_file, expected_result):
    # Act
    result = archiver.remove_file(archive_file)

    # Assert
    assert result == expected_result


@pytest.mark.parametrize(
    ("filename_lst", "expected_result"),
    [(["file1.txt", "file2.txt"], False), (["file3.txt"], False)],
    ids=["multiple_files", "single_file"],
)
def test_remove_files(archiver, filename_lst, expected_result):
    # Act
    result = archiver.remove_files(filename_lst)

    # Assert
    assert result == expected_result


def test_get_filename_list(archiver):
    # Act
    result = archiver.get_filename_list()

    # Assert
    assert result == []


@pytest.mark.parametrize(
    ("other_archive", "expected_result"),
    [(Archiver(Path("/other/path")), False)],
    ids=["other_archive"],
)
def test_copy_from_archive(archiver, other_archive, expected_result):
    # Act
    result = archiver.copy_from_archive(other_archive)

    # Assert
    assert result == expected_result


@pytest.mark.parametrize(
    ("input_path", "expected_path"),
    [
        (Path("/home/bpepple/test1"), Path("/home/bpepple/test1")),
        (Path("/home/bpepple/test2"), Path("/home/bpepple/test2")),
        (Path("/home/bpepple/test3"), Path("/home/bpepple/test3")),
        (Path(), Path()),  # edge case: empty path
        (Path("/"), Path("/")),  # edge case: root path
    ],
    ids=[
        "happy_path_test1",
        "happy_path_test2",
        "happy_path_test3",
        "edge_case_empty_path",
        "edge_case_root_path",
    ],
)
def test_archiver_path(input_path, expected_path):
    # Arrange
    archiver = Archiver(input_path)

    # Act
    result = archiver.path

    # Assert
    assert result == expected_path



"""Support for mixed digit/string type Issue field.

Class for handling the odd permutations of an 'issue number' that the
comics industry throws at us.
e.g.: "12", "12.1", "0", "-1", "5AU", "100-2"
"""

# Copyright 2012-2014 Anthony Beville

from __future__ import annotations


class IssueString:
    """
    Handles issue number strings by breaking them into numeric and suffix parts, and provides methods to convert to
    different formats.
    """

    def __init__(self: IssueString, text: str) -> None:
        # sourcery skip: remove-unnecessary-cast
        """
        Initializes an IssueString object by parsing the input text into numeric and suffix parts.

        Args:
            text (str): The issue number string to parse.

        Returns:
            None
        """

        # break up the issue number string into 2 parts: the numeric and suffix string.
        # (assumes that the numeric portion is always first)

        self.num: float | None = None
        self.suffix: str = ""

        if text is None:
            return

        if isinstance(text, int | float):
            text = str(text)

        if not text:
            return

        # skip the minus sign if it's first
        start: int = 1 if text[0] == "-" else 0
        # if it's still not numeric at start skip it
        if text[start].isdigit() or text[start] == ".":
            idx = self._find_split_point(text, start)
            idx = self._move_trailing_numeric_decimal_to_suffix(idx, text)
            idx = self._determine_if_number_after_minus_sign(idx, start)

            part1 = text[:idx]
            part2 = text[idx:]

            if part1 != "":
                self.num = float(part1)
            self.suffix = part2
        else:
            self.suffix = text

    @staticmethod
    def _move_trailing_numeric_decimal_to_suffix(
        idx: int,
        text: str,
    ) -> int:
        """
        Moves a trailing numeric decimal to the suffix if there is other content after it.

        Args:
            idx (int): The index of the decimal point.
            text (str): The text to process.

        Returns:
            int: The updated index after moving the decimal if necessary.
        """

        # move trailing numeric decimal to suffix (only if there is other junk after )
        if text[idx - 1] == "." and len(text) != idx:
            idx -= 1
        return idx

    @staticmethod
    def _determine_if_number_after_minus_sign(
        idx: int,
        start: int,
    ) -> int:
        """
        Determines if there is a numeric value after a minus sign, adjusting the index accordingly.

        Args:
            idx (int): The current index position.
            start (int): The starting index position.

        Returns:
            int: The adjusted index based on the presence of a numeric value after the minus sign.
        """

        # if there is no numeric after the minus, make the minus part of
        # the suffix
        return 0 if idx == 1 and start == 1 else idx

    @staticmethod
    def _find_split_point(text: str, start: int) -> int:
        """
        Finds the split point in a string where the numeric part ends.

        Args:
            text (str): The input string to search for the split point.
            start (int): The starting index for the search.

        Returns:
            int: The index where the numeric part ends.
        """

        # walk through the string, look for split point (the first non-numeric)
        decimal_count: int = 0
        for idx in range(start, len(text)):
            if text[idx] not in "0123456789.":
                break
            # special case: also split on second "."
            if text[idx] == ".":
                decimal_count += 1
                if decimal_count > 1:
                    break
        else:
            idx = len(text)

        return idx

    def as_string(self: IssueString, pad: int = 0) -> str:
        """
        Returns a string representation of the IssueString with left-side zero padding.

        Args:
            pad (int): The number of left-side zeroes to pad with.

        Returns:
            str: The IssueString as a string with zero padding.
        """

        # return the float, left side zero-padded, with suffix attached
        if self.num is None:
            return self.suffix

        negative: bool = self.num < 0

        num_f: float = abs(self.num)

        num_int = int(num_f)
        num_s = str(num_f) if float(num_int) != num_f else str(num_int)
        num_s += self.suffix

        length = len(str(num_int))
        padding: str = "0" * (pad - length) if length < pad else ""
        num_s = padding + num_s
        if negative:
            num_s = f"-{num_s}"

        return num_s

    def as_float(self: IssueString) -> float | None:
        """
        Returns the IssueString as a float value with no suffix.

        Returns:
            float | None: The IssueString as a float value, or None if the numeric part is not present.
        """

        if self.suffix == "½":
            return self.num + 0.5 if self.num is not None else 0.5
        return self.num

    def as_int(self: IssueString) -> int | None:
        """
        Returns the integer version of the float value in the IssueString.

        Returns:
            int | None: The integer value of the IssueString, or None if the numeric part is not present.
        """

        return None if self.num is None else int(self.num)



import tempfile
from pathlib import Path

import pytest

from darkseid.archivers.zip import ZipArchiver


@pytest.fixture()
def temp_zip_file():
    with tempfile.NamedTemporaryFile(suffix=".cbz", delete=False) as tmp:
        yield Path(tmp.name)
    tmp.close()
    Path(tmp.name).unlink(missing_ok=True)


@pytest.fixture()
def zip_archiver(temp_zip_file):
    return ZipArchiver(temp_zip_file)


@pytest.mark.parametrize(
    ("archive_file", "data"),
    [
        ("test.txt", "Hello, World!"),
        ("folder/test.txt", "Nested Hello!"),
        ("empty.txt", ""),
    ],
    ids=["simple_file", "nested_file", "empty_file"],
)
def test_write_file(zip_archiver, archive_file, data):
    # Act
    result = zip_archiver.write_file(archive_file, data)

    # Assert
    assert result is True
    assert archive_file in zip_archiver.get_filename_list()
    assert zip_archiver.read_file(archive_file).decode() == data


# TODO: Add a test for BadZipFile
@pytest.mark.parametrize(
    ("archive_file", "expected_exception"), [("nonexistent.zip", OSError)], ids=["nonexistent_file"]
)
def test_read_file_error(zip_archiver, archive_file, expected_exception):
    # Act & Assert
    with pytest.raises(expected_exception):
        zip_archiver.read_file(archive_file)


@pytest.mark.parametrize(
    ("archive_file", "data"), [("test.txt", "Hello, World!")], ids=["simple_file"]
)
def test_remove_file(zip_archiver, archive_file, data):
    # Arrange
    zip_archiver.write_file(archive_file, data)

    # Act
    result = zip_archiver.remove_file(archive_file)

    # Assert
    assert result is True
    assert archive_file not in zip_archiver.get_filename_list()


@pytest.mark.parametrize(
    ("filename_lst", "data"),
    [(["test1.txt", "test2.txt"], ["Hello, World!", "Another Hello!"])],
    ids=["multiple_files"],
)
def test_remove_files(zip_archiver, filename_lst, data):
    # Arrange
    for file, content in zip(filename_lst, data, strict=False):
        zip_archiver.write_file(file, content)

    # Act
    result = zip_archiver.remove_files(filename_lst)

    # Assert
    assert result is True
    for file in filename_lst:
        assert file not in zip_archiver.get_filename_list()


@pytest.mark.parametrize(
    ("archive_file", "data"), [("test.txt", "Hello, World!")], ids=["simple_file"]
)
def test_get_filename_list(zip_archiver, archive_file, data):
    # Arrange
    zip_archiver.write_file(archive_file, data)

    # Act
    filenames = zip_archiver.get_filename_list()

    # Assert
    assert archive_file in filenames


@pytest.mark.parametrize(
    ("exclude_list", "data"), [(["test.txt"], "Hello, World!")], ids=["exclude_file"]
)
def test_rebuild(zip_archiver, exclude_list, data):
    # Arrange
    zip_archiver.write_file("test.txt", data)
    zip_archiver.write_file("keep.txt", "Keep this file")

    # Act
    result = zip_archiver._rebuild(exclude_list)  # noqa: SLF001

    # Assert
    assert result is True
    assert "test.txt" not in zip_archiver.get_filename_list()
    assert "keep.txt" in zip_archiver.get_filename_list()


@pytest.mark.parametrize(
    ("other_archive_files", "data"), [(["test.txt"], ["Hello, World!"])], ids=["simple_copy"]
)
def test_copy_from_archive(zip_archiver, other_archive_files, data):
    # Arrange
    other_archive = ZipArchiver(Path(tempfile.NamedTemporaryFile(suffix=".cbz", delete=False).name))
    for file, content in zip(other_archive_files, data, strict=False):
        other_archive.write_file(file, content)

    # Act
    result = zip_archiver.copy_from_archive(other_archive)

    # Assert
    assert result is True
    for file in other_archive_files:
        assert file in zip_archiver.get_filename_list()
        assert zip_archiver.read_file(file).decode() == content



import pytest

from darkseid.issue_string import IssueString

float_test_values = {
    ("1½", 1.5),
    ("½", 0.5),
    ("0.5", 0.5),
    ("0", 0.0),
    ("1", 1.0),
    ("22.BEY", 22.0),
    ("22A", 22.0),
    ("22-A", 22.0),
}


@pytest.mark.parametrize(("issue", "expected"), float_test_values)
def test_float_strings(issue: str, expected: float) -> None:
    assert IssueString(issue).as_float() == expected


int_test_values = {("1", 1), ("1.5", 1), ("", None)}


@pytest.mark.parametrize(("issue", "expected"), int_test_values)
def test_issue_int(issue: str, expected: int) -> None:
    assert IssueString(issue).as_int() == expected


string_test_values = {
    ("1", "001", 3),
    ("1.MU", "001.MU", 3),
    ("-1", "-001", 3),
    ("Test", "Test", 0),
}


@pytest.mark.parametrize(("issue", "expected", "pad"), string_test_values)
def test_issue_string_monsters_unleashed(issue: str, expected: str, pad: int) -> None:
    assert IssueString(issue).as_string(pad) == expected
